import { ProductRepository } from '@app/common/modules/product/repositories';
import { ProjectRepository } from '@app/common/modules/project/repositiories';
import { NewTagRepository } from '@app/common/modules/tag/repositories/tag-repository';
import {
  Injectable,
  ConflictException,
  NotFoundException,
  HttpException,
  HttpStatus,
} from '@nestjs/common';
import { CreateTagDto } from '../dtos/tags.dto';
import { ProductEntity } from '@app/common/modules/product/entities';
import { ProjectEntity } from '@app/common/modules/project/entities';
import { SuccessResponseDto } from '@app/common/dto/success.response.dto';
import { BaseResponseDto } from '@app/common/dto/base.response.dto';
import { NewTagEntity } from '@app/common/modules/tag';
import { BulkCreateTagsDto, ProcessTagDto } from '../dtos';
import { In, QueryRunner } from 'typeorm';
import { ProjectParam } from '@app/common/dto/project-param.dto';

@Injectable()
export class TagService {
  constructor(
    private readonly tagRepository: NewTagRepository,
    private readonly productRepository: ProductRepository,
    private readonly projectRepository: ProjectRepository,
  ) {}

  async getTagsByProjectUuid(params: ProjectParam): Promise<BaseResponseDto> {
    const { projectUuid } = params;
    await this.getProjectByUuid(projectUuid);

    const tags = await this.tagRepository.find({
      where: { project: { uuid: projectUuid } },
      relations: ['product', 'project'],
    });

    return new SuccessResponseDto({
      message: `Tags retrieved successfully for project UUID: ${projectUuid}`,
      data: tags,
    });
  }

  async createTag(
    dto: CreateTagDto,
    param: ProjectParam,
  ): Promise<BaseResponseDto> {
    const { name, productUuid } = dto;

    const product = await this.getProductByUuid(productUuid);
    const project = await this.getProjectByUuid(param.projectUuid);

    await this.validateTagUniqueness(name, param.projectUuid);

    const tag = this.tagRepository.create({
      name,
      product,
      project,
    } as Partial<NewTagEntity>);

    await this.tagRepository.save(tag);

    return new SuccessResponseDto({
      message: `Tag created successfully`,
      data: tag,
    });
  }

  async processTags(
    tagDtos: ProcessTagDto[],
    projectUuid: string,
    queryRunner?: QueryRunner,
  ): Promise<NewTagEntity[]> {
    try {
      if (!tagDtos || tagDtos.length === 0) {
        return [];
      }

      const newTagDtos: CreateTagDto[] = [];
      const existingTagUuids: string[] = [];
      let fetchedExistingTags: NewTagEntity[] = [];
      const directlyFetchedTags: NewTagEntity[] = [];

      // Separate existing tag UUIDs and new tag DTOs
      for (const tagDto of tagDtos) {
        if (tagDto.uuid) {
          existingTagUuids.push(tagDto.uuid);
        } else {
          if (!tagDto.name || !tagDto.productUuid) {
            throw new HttpException(
              `Tag name or product UUID is missing`,
              HttpStatus.BAD_REQUEST,
            );
          }

          const existingTag = await queryRunner.manager.findOne(NewTagEntity, {
            where: {
              name: tagDto.name,
              product: { uuid: tagDto.productUuid },
              project: { uuid: projectUuid },
            },
            relations: ['product'],
          });

          if (!existingTag) {
            newTagDtos.push(tagDto);
          } else {
            directlyFetchedTags.push(existingTag);
          }
        }
      }

      // Fetch existing tags using UUIDs
      if (existingTagUuids.length > 0) {
        fetchedExistingTags = await (queryRunner
          ? queryRunner.manager.find(NewTagEntity, {
              where: {
                uuid: In(existingTagUuids),
                project: { uuid: projectUuid },
              },
              relations: ['product'],
            })
          : this.tagRepository.find({
              where: {
                uuid: In(existingTagUuids),
                project: { uuid: projectUuid },
              },
              relations: ['product'],
            }));
      }

      // Ensure all provided UUIDs exist in the database
      if (fetchedExistingTags.length !== existingTagUuids.length) {
        const foundUuids = new Set(fetchedExistingTags.map((tag) => tag.uuid));
        const missingUuids = existingTagUuids.filter(
          (uuid) => !foundUuids.has(uuid),
        );

        throw new HttpException(
          `Some provided tag UUIDs do not exist in the project: ${missingUuids.join(', ')}`,
          HttpStatus.NOT_FOUND,
        );
      }

      let newlyCreatedTags: NewTagEntity[] = [];

      if (newTagDtos.length > 0) {
        newlyCreatedTags = await this.bulkCreateTags(
          { projectUuid, tags: newTagDtos },
          queryRunner,
        );
      }

      // Combine all found and created tags
      const allTags = [
        ...fetchedExistingTags,
        ...newlyCreatedTags,
        ...directlyFetchedTags,
      ];

      return allTags;
    } catch (error) {
      console.error(error);
      throw new HttpException(
        error instanceof HttpException
          ? error.message
          : 'An unexpected error occurred while processing tags',
        error instanceof HttpException
          ? error.getStatus()
          : HttpStatus.INTERNAL_SERVER_ERROR,
      );
    }
  }

  async bulkCreateTags(
    dto: BulkCreateTagsDto,
    queryRunner?: QueryRunner,
  ): Promise<NewTagEntity[]> {
    try {
      const { projectUuid, tags } = dto;
      const newTags: NewTagEntity[] = [];

      const project = await this.getProjectByUuid(projectUuid);

      // Extract unique product UUIDs
      const productUuids = Array.from(
        new Set(tags.map((tag) => tag.productUuid)),
      );

      const productMap = await this.getProductMap(productUuids);

      // Fetch existing tag names for this project
      const existingTags = await this.tagRepository.find({
        where: { project: { uuid: projectUuid } },
        relations: ['product'],
      });

      // Convert existing tags into a Map for quick lookup
      const existingTagMap = new Map<string, string | null>();
      existingTags.forEach((tag) => {
        existingTagMap.set(tag.name, tag.product?.uuid || null);
      });

      for (const tag of tags) {
        const existingProductUuid = existingTagMap.get(tag.name);

        if (existingProductUuid) {
          if (existingProductUuid !== tag.productUuid) {
            throw new HttpException(
              `Tag "${tag.name}" already exists but is associated with a different product.`,
              HttpStatus.CONFLICT,
            );
          }
        } else {
          newTags.push(
            this.tagRepository.create({
              name: tag.name,
              product: productMap.get(tag.productUuid),
              project,
            }),
          );
        }
      }

      if (newTags.length > 0) {
        if (queryRunner) {
          await queryRunner.manager.save(NewTagEntity, newTags);
        } else {
          await this.tagRepository.save(newTags);
        }
      } else {
      }

      return newTags;
    } catch (error) {
      console.error(error);
      throw new HttpException(
        error instanceof HttpException
          ? error.message
          : 'An unexpected error occurred while creating tags',
        error instanceof HttpException
          ? error.getStatus()
          : HttpStatus.INTERNAL_SERVER_ERROR,
      );
    }
  }

  private async getProductByUuid(uuid: string): Promise<ProductEntity> {
    const product = await this.productRepository.findOne({ where: { uuid } });
    if (!product) {
      throw new NotFoundException(`Product with UUID ${uuid} not found.`);
    }
    return product;
  }

  private async getProjectByUuid(uuid: string): Promise<ProjectEntity> {
    const project = await this.projectRepository.findOne({ where: { uuid } });
    if (!project) {
      throw new NotFoundException(`Project with UUID ${uuid} not found.`);
    }
    return project;
  }

  private async validateTagUniqueness(
    name: string,
    projectUuid: string,
  ): Promise<void> {
    const existingTag = await this.tagRepository.findOne({
      where: { name, project: { uuid: projectUuid } },
    });

    if (existingTag) {
      throw new ConflictException(
        `Tag with name "${name}" already exists in this project.`,
      );
    }
  }

  private async getProductMap(
    uuids: string[],
  ): Promise<Map<string, ProductEntity>> {
    const products = await this.productRepository.find({
      where: { uuid: In(uuids) },
    });

    if (products.length !== uuids.length) {
      const foundUuids = new Set(products.map((p) => p.uuid));
      const missingUuids = uuids.filter((id) => !foundUuids.has(id));
      throw new NotFoundException(
        `Products not found for UUIDs: ${missingUuids.join(', ')}`,
      );
    }

    return new Map(products.map((product) => [product.uuid, product]));
  }

  private async getExistingTagNames(projectUuid: string): Promise<Set<string>> {
    const tags = await this.tagRepository.find({
      where: { project: { uuid: projectUuid } },
      select: ['name'],
    });

    return new Set(tags.map((tag) => tag.name));
  }
}
